#include "mainwindow.h"
#include "./ui_mainwindow.h"
#include "libusbwrapper.h"
#include <QMessageBox>
#include <memory>
#include <iostream>

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent)
    , ui(new Ui::MainWindow)
{
    ui->setupUi(this);
    connect(ui->rescanButton, SIGNAL(clicked()), this, SLOT(rescan_clicked()));
    connect(this, SIGNAL(close()), this, SLOT(on_close()));
    connect(ui->deviceComboBox, SIGNAL(currentTextChanged(QString)), this, SLOT(device_changed(QString)));
    connect(ui->clearKeycodesButton, SIGNAL(clicked()), this, SLOT(clear_keyboard_config()));
}

MainWindow::~MainWindow()
{
	delete ui;
}

void MainWindow::rescan_device_list()
{
	if (this->m_usb_devices.size()) {
		/* Close old devices */
		m_usb_devices.clear();
	}

	auto temp_vector = UsbDevice::find_devices(0xDEAD, 0xBEEF);
	this->m_usb_devices = std::vector<SustainPedalKeyboard>(temp_vector.begin(), temp_vector.end());

	ui->deviceComboBox->clear();

	/* Update combo box with devices */
	for (auto &dev : m_usb_devices) {
		dev.open();
		auto sn = dev.get_serial_number();
		ui->deviceComboBox->addItem(QString::fromStdString(sn));
	}

}

UsbDevice *MainWindow::get_device_from_sn(const std::string &sn)
{
	for (auto &d : m_usb_devices) {
		if (d.get_serial_number() == sn)
			return &d;
	}

	return NULL;
}

UsbHidKeyEvent MainWindow::create_key_event_pedal1()
{
	UsbHidKeyEvent event;

	event.left_control = ui->checkBoxLControl_1->isChecked();
	event.left_shift = ui->checkBoxLShift_1->isChecked();
	event.left_alt = ui->checkBoxLAlt_1->isChecked();
	event.left_super = ui->checkBoxLSuper_1->isChecked();
	event.right_control = ui->checkBoxRControl_1->isChecked();
	event.right_shift = ui->checkBoxRShift_1->isChecked();
	event.right_alt = ui->checkBoxRAlt_1->isChecked();
	event.right_super = ui->checkBoxRSuper_1->isChecked();

	event.keycodes[0] = (uint8_t)ui->keycode1_1->value();
	event.keycodes[1] = (uint8_t)ui->keycode2_1->value();
	event.keycodes[2] = (uint8_t)ui->keycode3_1->value();

	return event;
}

UsbHidKeyEvent MainWindow::create_key_event_pedal2()
{
	UsbHidKeyEvent event;

	event.left_control = ui->checkBoxLControl_2->isChecked();
	event.left_shift = ui->checkBoxLShift_2->isChecked();
	event.left_alt = ui->checkBoxLAlt_2->isChecked();
	event.left_super = ui->checkBoxLSuper_2->isChecked();
	event.right_control = ui->checkBoxRControl_2->isChecked();
	event.right_shift = ui->checkBoxRShift_2->isChecked();
	event.right_alt = ui->checkBoxRAlt_2->isChecked();
	event.right_super = ui->checkBoxRSuper_2->isChecked();

	event.keycodes[0] = (uint8_t)ui->keycode1_2->value();
	event.keycodes[1] = (uint8_t)ui->keycode2_2->value();
	event.keycodes[2] = (uint8_t)ui->keycode3_2->value();

	return event;
}

void MainWindow::rescan_clicked()
{
	rescan_device_list();
}

void MainWindow::device_changed(QString sn)
{
	auto dev = get_device_from_sn(sn.toStdString());
	if (m_currently_selected_dev.get()) {
		m_currently_selected_dev->set_active_indicator(false);
	}
	m_currently_selected_dev.reset();

	if (dev == NULL) {
		enable_pedal_gui_elements(false);
		return;
	}

	enable_pedal_gui_elements(true);

	m_currently_selected_dev = std::make_unique<SustainPedalKeyboard>(*dev);

	m_currently_selected_dev->set_active_indicator(true);
	update_gui_from_keyboard();
}

void MainWindow::clear_keyboard_config()
{
	if (!m_currently_selected_dev.get()) {
		QMessageBox::warning(this, "Error", "No device selected!");
		return;
	}

	UsbHidKeyEvent empty_key;

	if (m_currently_selected_dev->program_keyboard_keycode(0, empty_key)) {
		QMessageBox::warning(this, "Error", "Request to device failed!");
		return;
	}

	if (m_currently_selected_dev->program_keyboard_keycode(1, empty_key)) {
		QMessageBox::warning(this, "Error", "Request to device failed!");
		return;
	}
	update_gui_from_keyboard();
	QMessageBox::information(this, "Device Reset", "All configuration deleted!");

}

void MainWindow::update_gui_from_keyboard()
{
	UsbHidKeyEvent pedal1, pedal2;

	if (!m_currently_selected_dev->read_keyboard_config(&pedal1, &pedal2)) {
		ui->checkBoxLControl_1->setChecked(pedal1.left_control);
		ui->checkBoxLShift_1->setChecked(pedal1.left_shift);
		ui->checkBoxLAlt_1->setChecked(pedal1.left_alt);
		ui->checkBoxLSuper_1->setChecked(pedal1.left_super);
		ui->checkBoxRControl_1->setChecked(pedal1.right_control);
		ui->checkBoxRShift_1->setChecked(pedal1.right_shift);
		ui->checkBoxRAlt_1->setChecked(pedal1.right_alt);
		ui->checkBoxRSuper_1->setChecked(pedal1.right_super);

		ui->checkBoxLControl_2->setChecked(pedal2.left_control);
		ui->checkBoxLShift_2->setChecked(pedal2.left_shift);
		ui->checkBoxLAlt_2->setChecked(pedal2.left_alt);
		ui->checkBoxLSuper_2->setChecked(pedal2.left_super);
		ui->checkBoxRControl_2->setChecked(pedal2.right_control);
		ui->checkBoxRShift_2->setChecked(pedal2.right_shift);
		ui->checkBoxRAlt_2->setChecked(pedal2.right_alt);
		ui->checkBoxRSuper_2->setChecked(pedal2.right_super);

		ui->keycode1_1->setValue((int)pedal1.keycodes[0]);
		ui->keycode2_1->setValue((int)pedal1.keycodes[1]);
		ui->keycode3_1->setValue((int)pedal1.keycodes[2]);

		ui->keycode1_2->setValue((int)pedal2.keycodes[0]);
		ui->keycode2_2->setValue((int)pedal2.keycodes[1]);
		ui->keycode3_2->setValue((int)pedal2.keycodes[2]);
	}
}

void MainWindow::closeEvent(QCloseEvent *event)
{
	if (m_currently_selected_dev)
		m_currently_selected_dev->set_active_indicator(false);
	m_usb_devices.clear();
	m_currently_selected_dev.reset();
	m_currently_selected_dev = nullptr;
}

void MainWindow::showEvent(QShowEvent *event)
{
	QMainWindow::showEvent(event);
	enable_pedal_gui_elements(false);
	rescan_device_list();
}


void MainWindow::on_buttonProgPedal1_clicked()
{
	if (!m_currently_selected_dev.get())
		return;
	auto ev = create_key_event_pedal1();
	m_currently_selected_dev->program_keyboard_keycode(0, ev);
}


void MainWindow::on_buttonProgPedal2_clicked()
{
	if (!m_currently_selected_dev.get())
		return;

	auto ev = create_key_event_pedal2();
	m_currently_selected_dev->program_keyboard_keycode(1, ev);
}

void MainWindow::enable_pedal_gui_elements(bool enable)
{
	int count1 = ui->verticalLayoutPedal1->count();
	int count2 = ui->verticalLayoutPedal2->count();

	for (int i = 0; i < count1; i++) {
		ui->verticalLayoutPedal1->itemAt(i)->widget()->setEnabled(enable);
	}

	for (int i = 0; i < count2; i++) {
		ui->verticalLayoutPedal2->itemAt(i)->widget()->setEnabled(enable);
	}

	ui->clearKeycodesButton->setEnabled(enable);

}

