#include "sustainpedalkeyboard.h"
#include <iostream>

SustainPedalKeyboard::SustainPedalKeyboard(libusb_device *device) : UsbDevice(device)
{
	m_control_timeout = 500;
}

SustainPedalKeyboard::SustainPedalKeyboard(const UsbDevice &device) : UsbDevice(device)
{
	m_control_timeout = 500;
}

SustainPedalKeyboard::~SustainPedalKeyboard()
{
}

int SustainPedalKeyboard::get_control_timeout()
{
	return m_control_timeout;
}

void SustainPedalKeyboard::set_control_timeout(int timeout)
{
	m_control_timeout = timeout;
}

int SustainPedalKeyboard::set_led(uint8_t led, bool state)
{
	int res = this->control_transfer((1<<6), 0x1, state ? 1 : 0, led, NULL, 0, m_control_timeout);

	return res;
}

void SustainPedalKeyboard::set_active_indicator(bool state)
{
	(void)set_led(3, state);
}

int SustainPedalKeyboard::program_keyboard_keycode(uint8_t pedal_idx, UsbHidKeyEvent &key)
{
	uint32_t word;
	uint8_t data[4];

	if (pedal_idx > 1)
		return -1001;

	word = key.to_eeprom_word();

	for (int i = 0; i < 4; i++) {
		data[i] = word & 0xFF;
		word >>= 8;
	}

	int res = write_eeprom(pedal_idx * 4, data, 4);
	if (res < 0)
		return res;
	else
		return 0;
}

int SustainPedalKeyboard::read_keyboard_config(UsbHidKeyEvent *pedal1, UsbHidKeyEvent *pedal2)
{
	unsigned char eeprom_data[8];

	if (!pedal1 || !pedal2)
		return -1000;

	int res = read_eeprom(0, eeprom_data, sizeof(eeprom_data));
	if (res != 8) {
		return -1;
	}

	pedal1->from_eeprom_data(&eeprom_data[0]);
	pedal2->from_eeprom_data(&eeprom_data[4]);

	return 0;
}


int SustainPedalKeyboard::write_eeprom(uint16_t offset_addr, unsigned char *data, uint16_t len)
{
	return this->control_transfer((1<<6), 0x2, 0, offset_addr, data, len, m_control_timeout);
}

int SustainPedalKeyboard::read_eeprom(uint16_t offset_addr, unsigned char *data, uint16_t len)
{
	int ret;

	ret = this->control_transfer(0x80 | (1<<6), 0x2, 0x0, offset_addr, data, len, m_control_timeout);

	return ret;
}


UsbHidKeyEvent::UsbHidKeyEvent()
{
	for (int i = 0; i < 3; i++)
		keycodes[i] = 0;
	left_control = false;
	left_alt = false;
	left_shift = false;
	left_super = false;
	right_control = false;
	right_super = false;
	right_alt = false;
	right_shift = false;
}

uint32_t UsbHidKeyEvent::to_eeprom_word()
{
	uint8_t modifiers = 0;
	uint32_t ret;


	if (left_control)
		modifiers |= 1;
	if (left_shift)
		modifiers |= 2;
	if (left_alt)
		modifiers |= 4;
	if (left_super)
		modifiers |= 8;
	if (right_control)
		modifiers |= 16;
	if (right_shift)
		modifiers |= 32;
	if (right_alt)
		modifiers |= 64;
	if (right_super)
		modifiers |= 128;

	ret = (uint32_t)modifiers;
	ret |= (((uint32_t)keycodes[0]) << 8);
	ret |= (((uint32_t)keycodes[1]) << 16);
	ret |= (((uint32_t)keycodes[2]) << 24);

	return ret;
}

void UsbHidKeyEvent::from_eeprom_data(uint8_t data[4])
{
		left_control = !!(data[0] & 1);
		left_shift = !!(data[0] & 2);
		left_alt = !!(data[0] & 4);
		left_super = !!(data[0] & 8);
		right_control = !!(data[0] & 16);
		right_shift = !!(data[0] & 32);
		right_alt = !!(data[0] & 64);
		right_super = !!(data[0] & 128);

		for (int i = 0; i < 3; i++) {
			keycodes[i] = data[i + 1];
		}
}
